#!/usr/bin/env fades

import argparse
import hashlib
import logging
import pathlib
import sys
import time

import dateutil.parser  # fades python-dateutil
import httplib2  # fades
import humanize  # fades
import requests  # fades
import yaml  # fades
from apiclient import discovery  # fades google-api-python-client-py3
from oauth2client import client, tools  # fades
from oauth2client.file import Storage

# hack to fix 'apiclient' for modern Python
from collections.abc import Callable
import collections
collections.Callable = Callable

# info for auth
SCOPE = 'https://www.googleapis.com/auth/photoslibrary.readonly'
APPLICATION_NAME = 'Auto photo downloader'

# known mime types
IMAGE_MIMETYPE = "image/jpeg"
VIDEO_MIMETYPE = "video/mp4"

RETRY_ATTEMPTS = 3
DOWNLOADS_PER_RUN = 50
CHUNK_SIZE = 65536

# set up logging
logger = logging.getLogger("msg")
h = logging.StreamHandler()
h.setFormatter(logging.Formatter("%(asctime)s %(levelname)-12s %(message)s"))
logger.addHandler(h)
logger.setLevel(logging.INFO)


def get_credentials(auth_flags, config):
    """Gets valid user credentials from storage.

    If nothing has been stored, or if the stored credentials are invalid,
    the OAuth2 flow is completed to obtain the new credentials.

    Returns:
        Credentials, the obtained credential.
    """
    store = Storage(config['credentials_file'])
    credentials = store.get()
    if not credentials or credentials.invalid:
        flow = client.flow_from_clientsecrets(config['client_secret_file'], SCOPE)
        flow.user_agent = APPLICATION_NAME
        credentials = tools.run_flow(flow, store, auth_flags)
        logger.info('Storing credentials to %s', config['credentials_file'])
    return credentials


def retryable(orig_func):
    def new_func(*args, **kwargs):
        for attempt in range(1, RETRY_ATTEMPTS + 1):
            try:
                result = orig_func(*args, **kwargs)
            except Exception as exc:
                logger.error("Explorer failure in %s (attempt %d): %r", orig_func, attempt, exc)
                if attempt == RETRY_ATTEMPTS:
                    raise
                time.sleep(5)
            else:
                return result
    return new_func


class Explorer:
    def __init__(self, auth_flags, config):
        credentials = get_credentials(auth_flags, config)
        http = credentials.authorize(httplib2.Http())
        self.service = discovery.build('photoslibrary', 'v1', http=http)

    @retryable
    def list(self):
        """Get the active items in a folder."""
        media_items_service = self.service.mediaItems()
        request = media_items_service.list(pageSize=50)

        all_items = []
        while True:
            result = request.execute()
            media_items = result.get('mediaItems')
            if media_items is None:
                break
            all_items.extend(media_items)
            logger.debug("    adding remote content: %d", len(all_items))
            request = media_items_service.list_next(request, result)
            if request is None:
                break

        return all_items


def main(explorer, config):
    """Main entry point."""
    base_dest_dir = pathlib.Path(config['destination'])
    prefix = config['photo_prefix']

    logger.info("Listing remote content.")
    nodes = explorer.list()
    logger.info("    found %d nodes", len(nodes))
    to_download = []

    for node in nodes:
        node_id = node['id']
        mime_type = node['mimeType']
        base_url = node["baseUrl"]
        logger.debug("Node: %r (%s)", node_id, mime_type)

        created_at = dateutil.parser.parse(node["mediaMetadata"]["creationTime"])
        digest = hashlib.md5(node_id.encode("ascii")).hexdigest()[:5]

        base_name = f"{prefix}-{created_at:%Y%m%d-%H%M%S}-{digest}"

        # get the final filename from the node
        if mime_type == IMAGE_MIMETYPE:
            final_path = (base_dest_dir / base_name).with_suffix(".jpeg")
            download_indicator = "d"
        elif mime_type == VIDEO_MIMETYPE:
            final_path = (base_dest_dir / base_name).with_suffix(".mp4")
            download_indicator = "dv"
        else:
            raise ValueError(f"Unrecognized mimetype: {mime_type!r}")

        if final_path.exists():
            logger.debug("   (skipping, already downloaded)")
        else:
            to_download.append((final_path, base_url, download_indicator))

    # sort all items to download so we start from the oldest (specially important if we
    # trim the total to download)
    to_download.sort()

    logger.info("To download: %d", len(to_download))
    if len(to_download) > DOWNLOADS_PER_RUN:
        to_download = to_download[:DOWNLOADS_PER_RUN]
        logger.info("    (trimmed to first %d)", DOWNLOADS_PER_RUN)

    for final_path, base_url, download_indicator in to_download:
        final_url = f"{base_url}={download_indicator}"

        # download the file to a temporary name
        temp_path = final_path.with_suffix(final_path.suffix + ".temp")
        logger.debug("Downloading node to %s", final_path)

        response = requests.get(final_url)
        cl = response.headers.get("Content-Length")
        if cl is None:
            cl_nice = "?"
        else:
            cl_nice = humanize.naturalsize(int(cl), binary=True)
        logger.debug("    content length: %s", cl_nice)

        with temp_path.open("wb") as fh:
            for chunk in response.iter_content(CHUNK_SIZE):
                fh.write(chunk)
        temp_path.rename(final_path)

    logger.info("Done.")


if __name__ == '__main__':
    # manually handling sys.argv as then it's automatically processed by Explorer
    if '-v' in sys.argv:
        sys.argv.remove('-v')
        logger.setLevel(logging.DEBUG)
    if '-q' in sys.argv:
        sys.argv.remove('-q')
        logger.setLevel(logging.WARNING)

    config_file = sys.argv.pop(1)
    with open(config_file, "rt", encoding="utf8") as fh:
        config = yaml.safe_load(fh)

    logger.info("Init.")
    auth_flags = argparse.ArgumentParser(parents=[tools.argparser]).parse_args()
    explorer = Explorer(auth_flags, config)
    main(explorer, config)
